@@ -14,14 +14,18 @@ use strict;
 use vars qw($VERSION);
 use Image::ExifTool qw(:DataAccess :Utils);
 
-$VERSION = '1.01';
+$VERSION = '1.05';
 
 my %noYes = ( 0 => 'No', 1 => 'Yes' );
 
+# Matroska tags
+# Note: The tag ID's in the Matroska documentation include the length designation
+#       (the upper bits), which is not included in the tag ID's below
 %Image::ExifTool::Matroska::Main = (
     GROUPS => { 2 => 'Video' },
     NOTES => q{
-        Tags extracted from Matroska multimedia container (MKA, MKV and MKS) files.
+        The following tags are extracted from Matroska multimedia container files. 
+        This container format is used by file types such as MKA, MKV, MKS and WEBM. 
         For speed, ExifTool extracts tags only up to the first Cluster unless the
         Verbose (-v) or Unknown = 2 (-U) option is used.  See
         L<http://www.matroska.org/technical/specs/index.html> for the official
@@ -40,7 +44,12 @@ my %noYes = ( 0 => 'No', 1 => 'Yes' );
     0x2f7 => { Name => 'EBMLReadVersion',   Format => 'unsigned' },
     0x2f2 => { Name => 'EBMLMaxIDLength',   Format => 'unsigned', Unknown => 1 },
     0x2f3 => { Name => 'EBMLMaxSizeLength', Format => 'unsigned', Unknown => 1 },
-    0x282 => { Name => 'DocType',           Format => 'string' },
+    0x282 => {
+        Name => 'DocType',
+        Format => 'string',
+        # override FileType for "webm" files
+        RawConv => '$self->OverrideFileType("WEBM") if $val eq "webm"; $val',
+    },
     0x287 => { Name => 'DocTypeVersion',    Format => 'unsigned' },
     0x285 => { Name => 'DocTypeReadVersion',Format => 'unsigned' },
 #
@@ -377,7 +386,7 @@ my %noYes = ( 0 => 'No', 1 => 'Yes' );
         Unknown => 1,
         Groups => { 2 => 'Audio' },
     },
-    0x2264 => { Name => 'AudioSampleBits',  Format => 'unsigned', Groups => { 2 => 'Audio' } },
+    0x2264 => { Name => 'AudioBitsPerSample',   Format => 'unsigned', Groups => { 2 => 'Audio' } },
 #
 # Content Encoding
 #
@@ -627,7 +636,7 @@ my %noYes = ( 0 => 'No', 1 => 'Yes' );
 #------------------------------------------------------------------------------
 # Get variable-length Matroska integer
 # Inputs: 0) data buffer, 1) position in data
-# Returns: integer value and updates position, -1 for unknown value,
+# Returns: integer value and updates position, -1 for unknown/reserved value,
 #          or undef if no data left
 sub GetVInt($$)
 {
@@ -663,7 +672,7 @@ sub GetVInt($$)
 # Returns: 1 on success, 0 if this wasn't a valid Matroska file
 sub ProcessMKV($$)
 {
-    my ($exifTool, $dirInfo) = @_;
+    my ($et, $dirInfo) = @_;
     my $raf = $$dirInfo{RAF};
     my ($buff, $buf2, @dirEnd, $trackIndent, %trackTypes);
 
@@ -678,25 +687,25 @@ sub ProcessMKV($$)
     # verify header length
     my $hlen = GetVInt($buff, $pos);
     return 0 unless $hlen and $hlen > 0;
-    $pos + $hlen > $dataLen and $exifTool->Warn('Truncated Matroska header'), return 1;
-    $exifTool->SetFileType();
+    $pos + $hlen > $dataLen and $et->Warn('Truncated Matroska header'), return 1;
+    $et->SetFileType();
     SetByteOrder('MM');
     my $tagTablePtr = GetTagTable('Image::ExifTool::Matroska::Main');
 
     # set flag to process entire file (otherwise we stop at the first Cluster)
-    my $verbose = $exifTool->Options('Verbose');
-    my $processAll = ($verbose or $exifTool->Options('Unknown') > 1);
-    $$exifTool{TrackTypes} = \%trackTypes;  # store Track types reference
-    my $oldIndent = $$exifTool{INDENT};
+    my $verbose = $et->Options('Verbose');
+    my $processAll = ($verbose or $et->Options('Unknown') > 1);
+    $$et{TrackTypes} = \%trackTypes;  # store Track types reference
+    my $oldIndent = $$et{INDENT};
     my $chapterNum = 0;
 
     # loop over all Matroska elements
     for (;;) {
-        while (@dirEnd and $pos + $dataPos >= $dirEnd[-1]) {
+        while (@dirEnd and $pos + $dataPos >= $dirEnd[-1][0]) {
             pop @dirEnd;
             # use INDENT to decide whether or not we are done this Track element
-            delete $$exifTool{SET_GROUP1} if $trackIndent and $trackIndent eq $$exifTool{INDENT};
-            $$exifTool{INDENT} = substr($$exifTool{INDENT}, 0, -2);
+            delete $$et{SET_GROUP1} if $trackIndent and $trackIndent eq $$et{INDENT};
+            $$et{INDENT} = substr($$et{INDENT}, 0, -2);
         }
         # read more if we are getting close to the end of our buffer
         # (24 more bytes should be enough to read this element header)
@@ -710,32 +719,61 @@ sub ProcessMKV($$)
         my $tag = GetVInt($buff, $pos);
         last unless defined $tag and $tag >= 0;
         my $size = GetVInt($buff, $pos);
-        my $tagInfo = $exifTool->GetTagInfo($tagTablePtr, $tag);
+        last unless defined $size;
+        my $unknownSize;
+        $size < 0 and $unknownSize = 1, $size = 1e20;
+        if (@dirEnd and $pos + $dataPos + $size > $dirEnd[-1][0]) {
+            $et->Warn("Invalid or corrupted $dirEnd[-1][1] master element");
+            $pos = $dirEnd[-1][0] - $dataPos;
+            if ($pos < 0 or $pos > $dataLen) {
+                $buff = '';
+                $dataPos += $pos;
+                $dataLen = 0;
+                $pos = 0;
+                $raf->Seek($dataPos, 0) or last;
+            }
+            next;
+        }
+        my $tagInfo = $et->GetTagInfo($tagTablePtr, $tag);
         # just fall through into the contained EBML elements
         if ($tagInfo and $$tagInfo{SubDirectory}) {
             # stop processing at first cluster unless we are in verbose mode
             last if $$tagInfo{Name} eq 'Cluster' and not $processAll;
-            $$exifTool{INDENT} .= '| ';
-            $exifTool->VerboseDir($$tagTablePtr{$tag}{Name}, undef, $size);
-            push @dirEnd, $pos + $dataPos + $size;
+            $$et{INDENT} .= '| ';
+            $et->VerboseDir($$tagTablePtr{$tag}{Name}, undef, $size);
+            push @dirEnd, [ $pos + $dataPos + $size, $$tagInfo{Name} ];
             if ($$tagInfo{Name} eq 'ChapterAtom') {
-                $$exifTool{SET_GROUP1} = 'Chapter' . (++$chapterNum);
-                $trackIndent = $$exifTool{INDENT};
+                $$et{SET_GROUP1} = 'Chapter' . (++$chapterNum);
+                $trackIndent = $$et{INDENT};
             }
             next;
         }
-        last unless defined $size and $size >= 0;
+        last if $unknownSize;
         if ($pos + $size > $dataLen) {
-            # read more data in multiples of 64kB
-            my $more = (int(($pos + $size - $dataLen) / 65536) + 1) * 65536;
-            if ($raf->Read($buf2, $more)) {
-                $buff = substr($buff, $pos) . $buf2;
-                undef $buf2;
-                $dataPos += $pos;
-                $dataLen = length $buff;
+            # how much more do we need to read?
+            my $more = $pos + $size - $dataLen;
+            # just skip unknown and large data blocks
+            if (not $tagInfo or $more > 10000000) {
+                # don't try to skip very large blocks unless LargeFileSupport is enabled
+                last if $more > 0x80000000 and not $et->Options('LargeFileSupport');
+                $raf->Seek($more, 1) or last;
+                $buff = '';
+                $dataPos += $dataLen + $more;
+                $dataLen = 0;
                 $pos = 0;
+                next;
+            } else {
+                # read data in multiples of 64kB
+                $more = (int($more / 65536) + 1) * 65536;
+                if ($raf->Read($buf2, $more)) {
+                    $buff = substr($buff, $pos) . $buf2;
+                    undef $buf2;
+                    $dataPos += $pos;
+                    $dataLen = length $buff;
+                    $pos = 0;
+                }
+                last if $pos + $size > $dataLen;
             }
-            last if $pos + $size > $dataLen;
         }
         unless ($tagInfo) {
             # ignore the element
@@ -747,14 +785,14 @@ sub ProcessMKV($$)
             my $fmt = $$tagInfo{Format};
             if ($fmt eq 'string' or $fmt eq 'utf8') {
                 ($val = substr($buff, $pos, $size)) =~ s/\0.*//s;
-                $val = $exifTool->Decode($val, 'UTF8') if $fmt eq 'utf8';
+                $val = $et->Decode($val, 'UTF8') if $fmt eq 'utf8';
             } elsif ($fmt eq 'float') {
                 if ($size == 4) {
                     $val = GetFloat(\$buff, $pos);
                 } elsif ($size == 8) {
                     $val = GetDouble(\$buff, $pos);
                 } else {
-                    $exifTool->Warn("Illegal float size ($size)");
+                    $et->Warn("Illegal float size ($size)");
                 }
             } else {
                 my @vals = unpack("x${pos}C$size", $buff);
@@ -782,8 +820,8 @@ sub ProcessMKV($$)
             }
             # set group1 to Track/Chapter number
             if ($$tagInfo{Name} eq 'TrackNumber') {
-                $$exifTool{SET_GROUP1} = 'Track' . $val;
-                $trackIndent = $$exifTool{INDENT};
+                $$et{SET_GROUP1} = 'Track' . $val;
+                $trackIndent = $$et{INDENT};
             }
         }
         my %parms = (
@@ -793,20 +831,20 @@ sub ProcessMKV($$)
             Size    => $size,
         );
         if ($$tagInfo{NoSave}) {
-            $exifTool->VerboseInfo($tag, $tagInfo, Value => $val, %parms) if $verbose;
+            $et->VerboseInfo($tag, $tagInfo, Value => $val, %parms) if $verbose;
         } else {
-            $exifTool->HandleTag($tagTablePtr, $tag, $val, %parms);
+            $et->HandleTag($tagTablePtr, $tag, $val, %parms);
         }
         $pos += $size;  # step to next element
     }
-    $$exifTool{INDENT} = $oldIndent;
-    delete $$exifTool{SET_GROUP1};
+    $$et{INDENT} = $oldIndent;
+    delete $$et{SET_GROUP1};
     # override file type if necessary based on existing track types
     unless ($trackTypes{0x01} or $trackTypes{0x03}) {   # video or complex?
         if ($trackTypes{0x02}) {                        # audio?
-            $exifTool->OverrideFileType('MKA');
+            $et->OverrideFileType('MKA');
         } elsif ($trackTypes{0x11}) {                   # subtitle?
-            $exifTool->OverrideFileType('MKS');
+            $et->OverrideFileType('MKS');
         }
     }
     return 1;
@@ -827,11 +865,11 @@ This module is used by Image::ExifTool
 =head1 DESCRIPTION
 
 This module contains definitions required by Image::ExifTool to read meta
-information from Matroska multimedia files (MKA, MKV and MKS).
+information from Matroska multimedia files (MKA, MKV, MKS and WEBM).
 
 =head1 AUTHOR
 
-Copyright 2003-2010, Phil Harvey (phil at owl.phy.queensu.ca)
+Copyright 2003-2015, Phil Harvey (phil at owl.phy.queensu.ca)
 
 This library is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
